# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools


def _normalize_to_tuple(val):
    if isinstance(val, tuple):
        return val
    else:
        return (val,)


def parametrize(argnames, argvalues):
    """
    Decorate a test launch description in a way that causes it to run with specified parameters.

    This decorator behaves like the @pytest.mark.parametrize decorator.

    :param: argnames A comma separated list of argument names.

    :param: argvalues The values to use for arguments specified in argnames
    """
    argnames = [x.strip() for x in argnames.split(',') if x.strip()]
    argvalues = [_normalize_to_tuple(x) for x in argvalues]

    def _decorator(func):
        @functools.wraps(func)
        def _wrapped():
            for val in argvalues:
                partial_args = dict(zip(argnames, val))

                partial = functools.partial(func, **partial_args)
                functools.update_wrapper(partial, func)
                yield partial, partial_args
        _wrapped.__parametrized__ = True
        return _wrapped

    return _decorator



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import itertools
import os
import unittest
import warnings

from .actions import ReadyToTest


# Patch up the warnings module to streamline the warning messages.  See
# https://docs.python.org/3/library/warnings.html#warnings.showwarning
def slim_formatwarning(msg, *args, **kwargs):
    return 'Warning: ' + str(msg) + os.linesep


warnings.formatwarning = slim_formatwarning


def _normalize_ld(launch_description_fn):
    # A launch description fn can return just a launch description, or a tuple of
    # (launch_description, test_context).  This wrapper function normalizes things
    # so we always get a tuple, sometimes with an empty dictionary for the test_context
    def normalize(result):
        if isinstance(result, tuple):
            return result
        else:
            return result, {}

    def wrapper(**kwargs):
        # This is a new-style launch_description which should contain a ReadyToTest action
        ready_fn = kwargs.pop('ready_fn')
        result = normalize(launch_description_fn(**kwargs))
        # Fish the ReadyToTest action out of the launch description and plumb our
        # ready_fn to it

        def iterate_ready_to_test_actions(entities):
            """Recursively search LaunchDescription entities for all ReadyToTest actions."""
            for entity in entities:
                if isinstance(entity, ReadyToTest):
                    yield entity
                yield from iterate_ready_to_test_actions(
                    entity.describe_sub_entities()
                )
                for conditional_sub_entity in entity.describe_conditional_sub_entities():
                    yield from iterate_ready_to_test_actions(
                        conditional_sub_entity[1]
                    )

        try:
            ready_action = next(e for e in iterate_ready_to_test_actions(result[0].entities))
        except StopIteration:  # No ReadyToTest action found
            raise Exception(
                'generate_test_description functions must return '
                'a LaunchDescription containing a ReadyToTest action'
            )
        ready_action._add_callback(ready_fn)
        return result

    return wrapper


class TestRun:

    def __init__(self,
                 name,
                 test_description_function,
                 param_args,
                 pre_shutdown_tests,
                 post_shutdown_tests):
        self.name = name
        if not hasattr(test_description_function, '__markers__'):
            test_description_function.__markers__ = {}
        self._test_description_function = test_description_function
        self.normalized_test_description = _normalize_ld(test_description_function)

        self.param_args = param_args

        self.pre_shutdown_tests = pre_shutdown_tests
        self.post_shutdown_tests = post_shutdown_tests

        # If we're parametrized, extend the test names so we can tell more easily what
        # params they were run with
        if self.param_args:
            for tc in itertools.chain(
                _iterate_tests_in_test_suite(pre_shutdown_tests),
                _iterate_tests_in_test_suite(post_shutdown_tests)
            ):
                test_method = getattr(tc, tc._testMethodName)
                new_name = tc._testMethodName + self._format_params()
                setattr(tc, '_testMethodName', new_name)
                setattr(tc, new_name, test_method)

        # Disable cleanup of test cases once they are run
        for tc in itertools.chain(
            _iterate_test_suites(pre_shutdown_tests),
            _iterate_test_suites(post_shutdown_tests)
        ):
            tc._removeTestAtIndex = lambda *args, **kwargs: None

    @property
    def markers(self):
        return self._test_description_function.__markers__

    def bind(self, tests, injected_attributes={}, injected_args={}):
        """
        Bind injected_attributes and injected_args to tests.

        Injected Attributes can be accessed from a test as self.name
        Injected Arguments can be accessed as an argument if the test has an argument with a
        matching name
        """
        # Inject test attributes into the test as self.whatever.  This method of giving
        # objects to the test is pretty inferior to injecting them as arguments to the
        # test methods - we may deprecate this in favor of everything being an argument
        for name, value in injected_attributes.items():
            _give_attribute_to_tests(value, name, tests)

        # Give objects with matching names as arguments to tests.  This doesn't have the
        # weird scoping and name collision issues that the above method has.  In fact,
        # we give proc_info and proc_output to the tests as arguments too, so anything
        # you can do with test attributes can also be accomplished with test arguments
        _bind_test_args_to_tests(injected_args, tests)

    def get_launch_description(self):
        """
        Get just the launch description portion of the test_description.

        This should only be used for the purposes of introspecting the launch description.  The
        returned launch description is not meant to be launched
        """
        return self.normalized_test_description(ready_fn=lambda: None)[0]

    def all_cases(self):
        yield from _iterate_tests_in_test_suite(self.pre_shutdown_tests)
        yield from _iterate_tests_in_test_suite(self.post_shutdown_tests)

    def __str__(self):
        return self.name + self._format_params()

    def _format_params(self):
        if not self.param_args:
            return ''
        else:
            str_args = map(str, self.param_args.values())
            return '[{}]'.format(', '.join(str_args))


def LoadTestsFromPythonModule(module, *, name='launch_tests'):
    if not hasattr(module.generate_test_description, '__parametrized__'):
        normalized_test_description_func = (
            lambda: [(module.generate_test_description, {})]
        )
    else:
        normalized_test_description_func = module.generate_test_description

    # If our test description is parameterized, we'll load a set of tests for each
    # individual launch
    return [TestRun(name,
                    description,
                    args,
                    PreShutdownTestLoader().loadTestsFromModule(module),
                    PostShutdownTestLoader().loadTestsFromModule(module))
            for description, args in normalized_test_description_func()]


def PreShutdownTestLoader():
    return _make_loader(False)


def PostShutdownTestLoader():
    return _make_loader(True)


def _make_loader(load_post_shutdown):

    class _loader(unittest.TestLoader):
        """TestLoader selectively loads pre-shutdown or post-shutdown tests."""

        def loadTestsFromTestCase(self, testCaseClass):
            if getattr(testCaseClass, '__post_shutdown_test__', False) == load_post_shutdown:
                # Isolate test classes instances on a per parameterization basis
                cases = super(_loader, self).loadTestsFromTestCase(
                    type(testCaseClass.__name__, (testCaseClass,), {
                        '__module__': testCaseClass.__module__
                    })
                )
                return cases
            # Empty test suites will be ignored by the test runner
            return self.suiteClass()

    return _loader()


def _bind_test_args_to_tests(context, test_suite):
    # Look for tests that expect additional arguments and bind items from the context
    # to the tests
    for test in _iterate_tests_in_test_suite(test_suite):
        # Need to reach a little deep into the implementation here to get the test
        # method.  See unittest.TestCase
        test_method = getattr(test, test._testMethodName)
        # Replace the test with a functools.partial that has the arguments
        # provided by the test context already bound
        setattr(
            test,
            test._testMethodName,
            _partially_bind_matching_args(test_method, context)
        )

        test.setUp = _partially_bind_matching_args(
            test.setUp,
            context
        )

        test.tearDown = _partially_bind_matching_args(
            test.tearDown,
            context
        )

    for test_class in _iterate_test_classes_in_test_suite(test_suite):
        test_class.setUpClass = _partially_bind_matching_args(
            test_class.setUpClass,
            context
        )
        test_class.tearDownClass = _partially_bind_matching_args(
            test_class.tearDownClass,
            context
        )


def _partially_bind_matching_args(unbound_function, arg_candidates):
    function_args = inspect.signature(unbound_function).parameters
    # We only want to bind the part of the context matches the test args
    matching_args = {k: v for (k, v) in arg_candidates.items() if k in function_args}
    return functools.partial(unbound_function, **matching_args)


def _give_attribute_to_tests(data, attr_name, test_suite):

    def _warn_getter(self):
        if not hasattr(self, '__warned'):
            warnings.warn(
                'Automatically adding attributes like self.{0} '
                'to the test class will be deprecated in a future release.  '
                'Instead, add {0} to the test method argument list to '
                'access the test object you need'.format(attr_name)
            )
            setattr(self, '__warned', True)

        return data

    # The effect of this is that every test will have `self.attr_name` available to it so that
    # it can interact with ROS2 or the process exit coes, or IO or whatever data we want
    for cls in _iterate_test_classes_in_test_suite(test_suite):
        setattr(cls, attr_name, property(fget=_warn_getter))


def _iterate_test_classes_in_test_suite(test_suite):
    classes = []
    for t in _iterate_tests_in_test_suite(test_suite):
        if t.__class__ not in classes:
            classes.append(t.__class__)
            yield t.__class__


def _iterate_test_suites(test_suite):
    try:
        iter(test_suite)
    except TypeError:
        pass
    else:
        if isinstance(test_suite, unittest.TestSuite):
            yield test_suite
        for test in test_suite:
            yield from _iterate_test_suites(test)


def _iterate_tests_in_test_suite(test_suite):
    try:
        iter(test_suite)
    except TypeError:
        # Base case - test_suite is not iterable, so it must be an individual test method
        yield test_suite
    else:
        # Otherwise, it's a test_suite, or a list of individual test methods.  recurse
        for test in test_suite:
            yield from _iterate_tests_in_test_suite(test)



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import unittest

import ament_index_python

import launch
import launch.actions
import launch.substitutions

import launch_testing
import launch_testing.actions
import launch_testing.util

import pytest


@pytest.mark.launch_test
def generate_test_description():
    dut_process = launch.actions.ExecuteProcess(
        cmd=[
            sys.executable,
            os.path.join(
                ament_index_python.get_package_prefix('launch_testing'),
                'lib/launch_testing',
                'terminating_proc.py',
            ),

            # Arguments
            launch.substitutions.LaunchConfiguration('dut_arg')
        ],
    )

    return launch.LaunchDescription([

        # This argument can be passed into the test, and can be discovered by running
        # launch_test --show-args
        launch.actions.DeclareLaunchArgument(
            'dut_arg',
            default_value=['default'],
            description='Passed to the terminating process',
        ),

        dut_process,

        # In tests where all of the procs under tests terminate themselves, it's necessary
        # to add a dummy process not under test to keep the launch alive. launch_test
        # provides a simple launch action that does this:
        launch_testing.util.KeepAliveProc(),

        launch_testing.actions.ReadyToTest()
    ]), {'dut_process': dut_process}


class TestTerminatingProcessStops(unittest.TestCase):

    def test_proc_terminates(self, proc_info, dut_process):
        proc_info.assertWaitForShutdown(process=dut_process, timeout=10)


@launch_testing.post_shutdown_test()
class TestProcessOutput(unittest.TestCase):

    def test_ran_with_arg(self, dut_process):
        self.assertNotIn(
            'default',
            dut_process.process_details['cmd'],
            'Try running: launch_test test_with_args.test.py dut_arg:=arg'
        )

    def test_arg_printed_in_output(self, proc_output, test_args, dut_process):
        launch_testing.asserts.assertInStdout(
            proc_output,
            test_args['dut_arg'],
            dut_process
        )

    def test_default_not_printed(self, proc_output, dut_process):
        with self.assertRaises(AssertionError):
            launch_testing.asserts.assertInStdout(
                proc_output,
                'default',
                dut_process
            )



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the GTest action."""

from typing import List
from typing import Optional
from typing import Union

from launch import LaunchContext
from launch import SomeActionsType
from launch import SomeSubstitutionsType
from launch.action import Action
from launch.actions import ExecuteProcess
from launch.actions import OpaqueFunction
from launch.actions import TimerAction
from launch.event import Event
from launch.event_handlers import OnProcessExit


class Test(ExecuteProcess):
    """Action that runs a test."""

    def __init__(
        self,
        *,
        timeout: Optional[Union[float, SomeSubstitutionsType]] = None,
        kill_timeout: Union[float, SomeSubstitutionsType] = 5.0,
        **kwargs
    ) -> None:
        """
        Create a Test action.

        Many arguments are passed to :class:`launch.ExecuteProcess`, so
        see the documentation for the class for additional details.

        :param: timeout the test will be killed after timeout seconds.
        """
        super().__init__(**kwargs)
        self.__timeout = timeout
        self.__timer = None

    @property
    def timeout(self):
        """Getter for timeout."""
        return self.__timeout

    def __on_process_exit(self, event: Event, context: LaunchContext) -> Optional[SomeActionsType]:
        """On shutdown event."""
        if self.__timer:
            self.__timer.cancel()

    def execute(self, context: LaunchContext) -> Optional[List[Action]]:
        """
        Execute the action.

        Delegated to :meth:`launch.actions.ExecuteProcess.execute`.
        """
        actions = super().execute(context)
        if not self.__timeout:
            return actions
        self.__timer = TimerAction(
            period=self.__timeout,
            actions=[OpaqueFunction(
                function=self._shutdown_process,
                kwargs={'send_sigint': True})])
        on_process_exit_event = OnProcessExit(
                on_exit=self.__on_process_exit,
                target_action=self
            )
        context.register_event_handler(on_process_exit_event)
        if not actions:
            return [self.__timer]

        return actions.append(self.__timer)



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import pathlib
from unittest import TestCase

from _pytest._code.code import ReprFileLocation

import pytest

from ..loader import LoadTestsFromPythonModule
from ..test_runner import LaunchTestRunner


def _pytest_version_ge(major, minor=0, patch=0):
    """Return True if pytest version is >= the given version."""
    pytest_version = tuple([int(v) for v in pytest.__version__.split('.')])
    assert pytest_version
    return pytest_version >= (major, minor, patch)


class LaunchTestFailure(Exception):

    def __init__(self, message, results):
        super().__init__()
        self.message = message
        self.results = results

    def __str__(self):
        return self.message


class LaunchTestFailureRepr:
    """A `_pytest._code.code.ExceptionReprChain`-like object."""

    def __init__(self, failures):
        lines = [
            line
            for _, error_description in failures
            for line in error_description.splitlines()
        ]
        max_length = max(len(line) for line in lines) if lines else 3
        thick_sep_line = '=' * max_length
        thin_sep_line = '-' * max_length
        self._fulldescr = '\n' + '\n\n'.join([
            '\n'.join([
                thick_sep_line, f'FAIL: {test_name}',
                thin_sep_line, error_description
            ])
            for test_name, error_description in failures
        ])
        self.reprcrash = ReprFileLocation(
            path='', lineno=0, message='\n'.join([
                f'{test_name} failed' for test_name, _ in failures
            ])
        )

    def __str__(self):
        return self._fulldescr

    def toterminal(self, out):
        out.write(self._fulldescr)


class LaunchTestItem(pytest.Item):

    def __init__(self, parent, *, name):
        super().__init__(name, parent)
        self.test_runs = None
        self.runner_cls = None

    @classmethod
    def from_parent(
        cls, parent, *, name, test_runs, runner_cls=LaunchTestRunner
    ):
        """Override from_parent for compatibility."""
        # pytest.Item.from_parent didn't exist before pytest 5.4
        if hasattr(super(), 'from_parent'):
            instance = getattr(super(), 'from_parent')(parent, name=name)
        else:
            instance = cls(parent, name=name)
        instance.test_runs = test_runs
        instance.runner_cls = runner_cls
        return instance

    def runtest(self):
        launch_args = sum((
            args_set for args_set in self.config.getoption('--launch-args')
        ), [])
        # Copy test runs' collection as it may be used more than
        # once e.g. if pytest rerunfailures plugin is in use.
        test_runs = copy.deepcopy(self.test_runs)
        runner = self.runner_cls(
            test_runs=test_runs,
            launch_file_arguments=launch_args,
            debug=self.config.getoption('verbose')
        )

        runner.validate()
        results_per_run = runner.run()

        if any(not result.wasSuccessful() for result in results_per_run.values()):
            raise LaunchTestFailure(
                message='some test cases have failed', results=results_per_run
            )

    def repr_failure(self, excinfo):
        if isinstance(excinfo.value, LaunchTestFailure):
            return LaunchTestFailureRepr(failures=[
                (test_case.id(), formatted_error)
                for test_run, test_result in excinfo.value.results.items()
                for test_case, formatted_error in (test_result.errors + test_result.failures)
                if isinstance(test_case, TestCase) and not test_result.wasSuccessful()
            ])
        return super().repr_failure(excinfo)

    def reportinfo(self):
        if _pytest_version_ge(7):
            path = self.path
        else:
            path = self.fspath
        return path, 0, 'launch tests: {}'.format(self.name)


class LaunchTestModule(pytest.File):

    def __init__(self, *args, **kwargs):
        if _pytest_version_ge(7):
            if 'fspath' in kwargs:
                if kwargs['fspath'] is not None:
                    kwargs['path'] = pathlib.Path(kwargs['fspath'])
                del kwargs['fspath']
        super().__init__(*args, **kwargs)

    @classmethod
    def from_parent(cls, *args, **kwargs):
        """Override from_parent for compatibility."""
        # pytest.File.from_parent didn't exist before pytest 5.4
        if _pytest_version_ge(5, 4):
            return super().from_parent(*args, **kwargs)
        args_without_parent = args[1:]
        return cls(*args_without_parent, **kwargs)

    def makeitem(self, *args, **kwargs):
        return LaunchTestItem.from_parent(*args, **kwargs)

    def collect(self):
        if _pytest_version_ge(7):
            # self.path exists since 7
            from _pytest.pathlib import import_path
            module = import_path(self.path, root=None)
        else:
            module = self.fspath.pyimport()
        yield self.makeitem(
            name=module.__name__, parent=self,
            test_runs=LoadTestsFromPythonModule(
                module, name=module.__name__
            )
        )


def find_launch_test_entrypoint(path):
    try:
        if _pytest_version_ge(7):
            from _pytest.pathlib import import_path
            module = import_path(path, root=None)
        else:
            # Assume we got legacy path in earlier versions of pytest
            module = path.pyimport()
        return getattr(module, 'generate_test_description', None)
    except SyntaxError:
        return None


def pytest_pycollect_makemodule(path, parent):
    entrypoint = find_launch_test_entrypoint(path)
    if entrypoint is not None:
        ihook = parent.session.gethookproxy(path)
        module = ihook.pytest_launch_collect_makemodule(
            path=path, parent=parent, entrypoint=entrypoint
        )
        if module is not None:
            return module

    if _pytest_version_ge(7):
        path = pathlib.Path(path)
        if path.name == '__init__.py':
            return pytest.Package.from_parent(parent, path=path)
        return pytest.Module.from_parent(parent=parent, path=path)
    elif _pytest_version_ge(5, 4):
        if path.basename == '__init__.py':
            return pytest.Package.from_parent(parent, fspath=path)
        return pytest.Module.from_parent(parent, fspath=path)
    else:
        # todo: remove fallback once all platforms use pytest >=5.4
        if path.basename == '__init__.py':
            return pytest.Package(path, parent)
        return pytest.Module(path, parent)


@pytest.hookimpl(trylast=True)
def pytest_launch_collect_makemodule(path, parent, entrypoint):
    marks = getattr(entrypoint, 'pytestmark', [])
    if marks and any(m.name == 'launch_test' for m in marks):
        if _pytest_version_ge(7):
            path = pathlib.Path(path)
            return LaunchTestModule.from_parent(parent=parent, path=path)
        else:
            return LaunchTestModule.from_parent(parent=parent, fspath=path)


def pytest_addhooks(pluginmanager):
    from launch_testing.pytest import hookspecs
    pluginmanager.add_hookspecs(hookspecs)


def pytest_addoption(parser):
    parser.addoption(
        '--launch-args', action='append', nargs='*',
        default=[], help='One or more Launch test arguments'
    )


def pytest_configure(config):
    config.addinivalue_line(
        'markers',
        'launch_test: mark a generate_test_description function as a launch test entrypoint'
    )



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import threading
import unittest

import launch
from launch import LaunchDescription
from launch import LaunchService
from launch.actions import RegisterEventHandler
from launch.event_handlers import OnProcessExit
from launch.event_handlers import OnProcessIO
from launch.event_handlers import OnProcessStart

from .io_handler import ActiveIoHandler
from .parse_arguments import parse_launch_arguments
from .proc_info_handler import ActiveProcInfoHandler
from .test_result import FailResult, SkipResult, TestResult


class _LaunchDiedException(Exception):
    pass


class _RunnerWorker():

    def __init__(self,
                 test_run,
                 test_run_preamble,
                 launch_file_arguments=[],
                 debug=False):
        self._test_run = test_run
        self._test_run_preamble = test_run_preamble
        self._launch_service = LaunchService(debug=debug)
        self._processes_launched = threading.Event()  # To signal when all processes started
        self._tests_completed = threading.Event()  # To signal when all the tests have finished
        self._launch_file_arguments = launch_file_arguments

        # Can't run LaunchService.run on another thread :-(
        # See https://github.com/ros2/launch/issues/126
        #
        # It would be simpler if we could run the pre-shutdown test and the post-shutdown tests on
        # one thread, and run the launch on another thead.
        #
        # Instead, we'll run the pre-shutdown tests on a background thread concurrent with the
        # launch on the main thread.  Once the launch is stopped, we'll run the post-shutdown
        # tests on the main thread
        self._test_tr = threading.Thread(
            target=self._run_test,
            name='test_runner_thread',
            daemon=True
        )

    def run(self):
        """
        Launch the processes under test and run the tests.

        :return: A tuple of two unittest.Results - one for tests that ran while processes were
        active, and another set for tests that ran after processes were shutdown
        """
        test_ld, test_context = self._test_run.normalized_test_description(
            ready_fn=lambda: self._processes_launched.set()
        )

        # Data that needs to be bound to the tests:
        proc_info = ActiveProcInfoHandler()
        proc_output = ActiveIoHandler()
        full_context = dict(test_context, **self._test_run.param_args)
        parsed_launch_arguments = parse_launch_arguments(self._launch_file_arguments)
        test_args = {}

        for k, v in parsed_launch_arguments:
            test_args[k] = v

        self._test_run.bind(
            self._test_run.pre_shutdown_tests,
            injected_attributes={
                'launch_service': self._launch_service,
                'proc_info': proc_info,
                'proc_output': proc_output,
                'test_args': test_args,
            },
            injected_args=dict(
                full_context,
                # Add a few more things to the args dictionary:
                **{
                    'launch_service': self._launch_service,
                    'proc_info': proc_info,
                    'proc_output': proc_output,
                    'test_args': test_args
                }
            )
        )
        self._test_run.bind(
            self._test_run.post_shutdown_tests,
            injected_attributes={
                'proc_info': proc_info._proc_info_handler,
                'proc_output': proc_output._io_handler,
                'test_args': test_args,
            },
            injected_args=dict(
                full_context,
                # Add a few more things to the args dictionary:
                **{
                    'proc_info': proc_info._proc_info_handler,
                    'proc_output': proc_output._io_handler,
                    'test_args': test_args
                }
            )
        )

        # Wrap the test_ld in another launch description so we can bind command line arguments to
        # the test and add our own event handlers for process IO and process exit:
        launch_description = LaunchDescription([
            *self._test_run_preamble,
            RegisterEventHandler(
                OnProcessStart(on_start=lambda info, unused: proc_info.append(info))
            ),
            RegisterEventHandler(
                OnProcessStart(
                    on_start=lambda info, unused: proc_output.track(info.process_name)
                )
            ),
            RegisterEventHandler(
                OnProcessExit(on_exit=lambda info, unused: proc_info.append(info))
            ),
            RegisterEventHandler(
                OnProcessIO(
                    on_stdout=proc_output.append,
                    on_stderr=proc_output.append,
                )
            ),
            launch.actions.IncludeLaunchDescription(
                launch.LaunchDescriptionSource(launch_description=test_ld),
                launch_arguments=parsed_launch_arguments
            ),
        ])

        self._launch_service.include_launch_description(
            launch_description
        )

        self._test_tr.start()  # Run the tests on another thread

        # This will block until the test thread stops it
        self._launch_service.run(
            shutdown_when_idle=not self._test_run.markers.get('keep_alive', False)
        )

        if not self._tests_completed.wait(timeout=0):
            # LaunchService.run returned before the tests completed.  This can be because the user
            # did ctrl+c, or because all of the launched nodes died before the tests completed
            print('Processes under test stopped before tests completed')
            # Give some extra help debugging why processes died early
            self._print_process_output_summary(proc_info, proc_output)
            # We treat this as a test failure and return some test results indicating such
            raise _LaunchDiedException()

        inactive_results = unittest.TextTestRunner(
            verbosity=2,
            resultclass=TestResult
        ).run(self._test_run.post_shutdown_tests)

        self._results.append(inactive_results)

        return self._results

    def _run_test(self):
        # Waits for the DUT processes to start (signaled by the _processes_launched
        # event) and then runs the tests

        if not self._processes_launched.wait(timeout=15):
            # Timed out waiting for the processes to start
            print('Timed out waiting for processes to start up')
            self._launch_service.shutdown()
            return

        try:
            # Run the tests
            self._results = unittest.TextTestRunner(
                verbosity=2,
                resultclass=TestResult
            ).run(self._test_run.pre_shutdown_tests)

        finally:
            self._tests_completed.set()
            self._launch_service.shutdown()

    def _print_process_output_summary(self, proc_info, proc_output):
        failed_procs = [proc for proc in proc_info if proc.returncode != 0]

        for process in failed_procs:
            print("Process '{}' exited with {}".format(process.process_name, process.returncode))
            print("##### '{}' output #####".format(process.process_name))
            try:
                for io in proc_output[process.action]:
                    print('{}'.format(io.text.decode('ascii')))
            except KeyError:
                pass  # Process generated no output
            print('#' * (len(process.process_name) + 21))


class LaunchTestRunner(object):

    def __init__(self,
                 test_runs,
                 launch_file_arguments=[],
                 debug=False):
        """
        Create an LaunchTestRunner object.

        :param callable gen_launch_description_fn: A function that returns a ros2 LaunchDesription
        for launching the processes under test.  This function should take a callable as a
        parameter which will be called when the processes under test are ready for the test to
        start
        """
        self._test_runs = test_runs
        self._launch_file_arguments = launch_file_arguments
        self._debug = debug

    def generate_preamble(self):
        """Generate a launch description preamble for a test to be run with."""
        return []

    def run(self):
        """
        Launch the processes under test and run the tests.

        :return: A tuple of two unittest.Results - one for tests that ran while processes were
        active, and another set for tests that ran after processes were shutdown
        """
        # We will return the results as a {test_run: TestResult)}
        results = {}

        for run in self._test_runs:
            if len(self._test_runs) > 1:
                print('\nStarting test run {}'.format(run))
            try:
                worker = _RunnerWorker(
                    run,
                    self.generate_preamble(),
                    self._launch_file_arguments,
                    self._debug)
                results[run] = worker.run()
            except unittest.case.SkipTest as skip_exception:
                # If a 'skip' decorator was placed on the generate_test_description function,
                # we skip all the tests for that run
                print('{} is skipped: {}'.format(run, skip_exception))
                results[run] = SkipResult(test_run=run, skip_reason=str(skip_exception))
                continue
            except _LaunchDiedException:
                # The most likely cause was ctrl+c, so we'll abort the test run
                results[run] = FailResult(
                    test_run=run,
                    message='Launch stopped before the active tests finished.'
                )
                break

        return results

    def validate(self):
        """Inspect the test configuration for configuration errors."""
        # Make sure the function signature of the launch configuration
        # generator is correct
        for run in self._test_runs:
            # Drill down into any parametrized test descriptions and make sure the argument names
            # are correct.  A simpler check can use getcallargs, but then you won't get a very
            # helpful message.
            base_fn = inspect.unwrap(run._test_description_function)
            base_args = inspect.getfullargspec(base_fn)
            base_args = base_args.args + base_args.kwonlyargs

            # Check that the parametrized arguments all have a place to go
            for argname in run.param_args.keys():
                if argname not in base_args:
                    raise Exception(
                        'Could not find an argument in generate_test_description matching '
                        "prametrized argument '{}'".format(argname)
                    )

            # Check for extra args in generate_test_description
            for argname in base_args:
                if argname not in run.param_args.keys():
                    raise Exception(
                        "generate_test_description has unexpected extra argument '{}'".format(
                            argname
                        )
                    )

            # This is a double-check
            try:
                inspect.getcallargs(run._test_description_function, ready_fn=lambda: None)
            except TypeError:
                # We also support generate_test_description functions without a ready_fn
                inspect.getcallargs(run._test_description_function)



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List
from typing import Optional

from launch.action import Action
from launch.launch_context import LaunchContext
from launch.launch_description_entity import LaunchDescriptionEntity

_logger_ = logging.getLogger(__name__)


class ReadyToTest(Action):
    """Action that signals to launch_test that it's safe to start the tests."""

    def __init__(self):
        super().__init__()
        self._cb_list = []

    def _add_callback(self, callback):
        self._cb_list.append(callback)

    def execute(self, context: LaunchContext) -> Optional[List[LaunchDescriptionEntity]]:
        """Execute the action."""
        for cb in self._cb_list:
            try:
                cb()
            except Exception as e:
                _logger_.error(e)



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib

import launch
import launch.actions
import launch.events
from launch.utilities import ensure_argument_type

from ..io_handler import ActiveIoHandler
from ..proc_info_handler import ActiveProcInfoHandler


class ProcessProxy:
    """A proxy to interact with `launch.actions.ExecuteProcess` instances."""

    def __init__(self, process_action, proc_info, proc_output, *, output_filter=None):
        """
        Construct a proxy for the given ``process_action``.

        :param process_action: `launch.actions.ExecuteProcess` instance to proxy.
        :param proc_info: `ActiveProcInfoHandler` tracking process state.
        :param proc_output: `ActiveIoHandler` tracking process output.
        :param output_filter: an optional callable to filter output text.
        """
        ensure_argument_type(
            process_action, types=launch.actions.ExecuteProcess, argument_name='process_action'
        )
        ensure_argument_type(proc_info, types=ActiveProcInfoHandler, argument_name='proc_info')
        ensure_argument_type(proc_output, types=ActiveIoHandler, argument_name='proc_output')
        if output_filter is not None and not callable(output_filter):
            raise TypeError(
                "Expected 'output_filter' to be callable but got '{!r}'".format(output_filter)
            )
        self._process_action = process_action
        self._proc_info = proc_info
        self._proc_output = proc_output
        self._output_filter = output_filter

    def wait_for_shutdown(self, timeout=None):
        """
        Wait for the target process to shutdown.

        :param timeout: time in seconds to wait, or None to block indefinitely.
        :return: whether the target process shut down or not.
        """
        with self._proc_info.proc_event:
            return self._proc_info.proc_event.wait_for(
                lambda: self.terminated, timeout=timeout
            )

    def wait_for_output(self, condition=None, timeout=None):
        """
        Wait for the target process to produce any output, either over stdout or stderr.

        :param condition: a callable to wait on a specific output condition to be satisfied,
            or ``None`` (default) to wake on any output.
        :param timeout: time in seconds to wait, or ``None`` (default) to block indefinitely.
        :return: whether the condition has been satisfied or not.
        """
        if condition is None:
            condition = (lambda output: True)
        if not callable(condition):
            raise TypeError(
                "Expected 'condition' to be callable but got '{!r}'".format(condition)
            )

        actual_output = None

        def remember_output():
            nonlocal actual_output
            actual_output = self.output
            return actual_output

        class BoolWithText:

            def __init__(self, result, output):
                self._result = result
                self._output = output

            def __bool__(self):
                return self._result

            def __repr__(self):
                return f'<BoolWithText({self._result}): {repr(self._output)}>'

        with self._proc_output.io_event:
            bool_result = self._proc_output.io_event.wait_for(
                lambda: self.running and condition(remember_output()), timeout=timeout
            )
            return BoolWithText(bool_result, actual_output)

    @property
    def target_process_action(self):
        return self._process_action

    @property
    def stderr(self):
        output_events = self._proc_output[self._process_action]
        output_text = ''.join(ev.text.decode() for ev in output_events if ev.from_stderr)
        if self._output_filter is not None:
            output_text = self._output_filter(output_text)
        return output_text

    @property
    def stdout(self):
        output_events = self._proc_output[self._process_action]
        output_text = ''.join(ev.text.decode() for ev in output_events if ev.from_stdout)
        if self._output_filter is not None:
            output_text = self._output_filter(output_text)
        return output_text

    @property
    def output(self):
        output_events = self._proc_output[self._process_action]
        output_text = ''.join(ev.text.decode() for ev in output_events)
        if self._output_filter is not None:
            output_text = self._output_filter(output_text)
        return output_text

    @property
    def running(self):
        if self._process_action not in self._proc_info.processes():
            return False
        return isinstance(
            self._proc_info[self._process_action],
            launch.events.process.ProcessStarted
        )

    @property
    def terminated(self):
        if self._process_action not in self._proc_info.processes():
            return False
        return isinstance(
            self._proc_info[self._process_action],
            launch.events.process.ProcessExited
        )

    @property
    def exit_code(self):
        return self._proc_info[self._process_action].returncode


@contextlib.contextmanager
def launch_process(launch_service, process_action, proc_info, proc_output, **kwargs):
    """
    Launch and interact with a process.

    On context entering, start execution of a ``process_action`` using the given ``launch_service``
    and yield a `ProcessProxy` to that ``process_action``.
    On context exiting, shut the process down if it has not been terminated yet.
    All additional arguments are forwarded to `ProcessProxy` on construction.
    """
    ensure_argument_type(
        process_action, types=launch.actions.ExecuteProcess, argument_name='process_action'
    )

    launch_service.emit_event(
        event=launch.events.IncludeLaunchDescription(
            launch_description=launch.LaunchDescription([process_action])
        )
    )
    process_proxy = ProcessProxy(process_action, proc_info, proc_output, **kwargs)
    try:
        yield process_proxy
    finally:
        if not process_proxy.terminated:
            launch_service.emit_event(
                event=launch.events.process.ShutdownProcess(
                    process_matcher=launch.events.matches_action(process_action)
                )
            )



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for the GTest action."""

import sys

from launch import SomeSubstitutionsType

from .test import Test


class PyTest(Test):
    """Action that runs a Pytest test."""

    def __init__(
        self,
        *,
        path: SomeSubstitutionsType,
        **kwargs
    ) -> None:
        """
        Create a PyTest test action.

        timeout argument is passed to :class:`launch_testing.Test`.
        The other arguments are passed to :class:`launch.ExecuteProcess`, so
        see the documentation for the class for additional details.

        :param: path to the test to be executed.
        """
        cmd = [sys.executable, '-m', 'pytest', path]
        super().__init__(cmd=cmd, **kwargs)
        self.__path = path

    @property
    def path(self):
        """Getter for path."""
        return self.__path



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time
import unittest


class FailResult(unittest.TestResult):
    """For test runs that fail when the DUT dies unexpectedly."""

    def __init__(self, test_run, message):
        super().__init__()
        for case in test_run.all_cases():
            self.addFailure(case, (Exception, Exception(message), None))
            self.testsRun += 1

    @property
    def testCases(self):
        return [failure[0] for failure in self.failures]

    @property
    def testTimes(self):
        """Get a dict of {test_case: elapsed_time}."""
        return {failure[0]: 0 for failure in self.failures}

    def wasSuccessful(self):
        return False


class SkipResult(unittest.TestResult):
    """For test runs with a skip decorator on the generate_test_description function."""

    def __init__(self, test_run, skip_reason=''):
        super().__init__()
        for case in test_run.all_cases():
            self.addSkip(case, skip_reason)
            self.testsRun += 1

    @property
    def testCases(self):
        return [skipped[0] for skipped in self.skipped]

    @property
    def testTimes(self):
        """Get a dict of {test_case: elapsed_time}."""
        return {skipped[0]: 0 for skipped in self.skipped}

    def wasSuccessful(self):
        return True


class TestResult(unittest.TextTestResult):
    """
    Subclass of unittest.TestResult that collects more information about the tests that ran.

    This class extends TestResult by recording all of the tests that ran, and by recording
    start and stop time for the individual test cases
    """

    def __init__(self, stream=None, descriptions=None, verbosity=None):
        self.__test_cases = {}
        super().__init__(stream, descriptions, verbosity)

    @property
    def testCases(self):
        return self.__test_cases.keys()

    @property
    def testTimes(self):
        """Get a dict of {test_case: elapsed_time}."""
        return {k: v['end'] - v['start'] for (k, v) in self.__test_cases.items()}

    def startTest(self, test):
        self.__test_cases[test] = {
            'start': time.time(),
            'end': 0
        }
        super().startTest(test)

    def stopTest(self, test):
        self.__test_cases[test]['end'] = time.time()
        super().stopTest(test)

    def append(self, results):
        self.__test_cases.update(results.__test_cases)

        self.failures += results.failures
        self.errors += results.errors
        self.testsRun += results.testsRun
        self.skipped += results.skipped
        self.expectedFailures += results.expectedFailures
        self.unexpectedSuccesses += results.unexpectedSuccesses



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module providing exit code assertions.

To prevent pytest from rewriting this module assertions, please PYTEST_DONT_REWRITE.
See https://docs.pytest.org/en/latest/assert.html#disabling-assert-rewriting for
further reference.
"""


import os

from ..util import resolveProcesses

EXIT_OK = 0
EXIT_SIGINT = 130
EXIT_SIGQUIT = 131
EXIT_SIGKILL = 137
EXIT_SIGSEGV = 139

# TODO(hidmic): Drop when SIGINT is fixed on Windows
EXIT_FORCED = 1


def assertExitCodes(proc_info,
                    allowable_exit_codes=[EXIT_OK] if os.name != 'nt' else [EXIT_OK, EXIT_FORCED],
                    process=None,  # By default, checks all processes
                    cmd_args=None,
                    *,
                    strict_proc_matching=True):
    """
    Check the exit codes of the processes under test.

    :param iterable proc_info: A list of proc_info objects provided by the test framework to be
    checked
    """
    # Sanity check that the user didn't pass in something crazy for allowable exit codes
    for code in allowable_exit_codes:
        assert isinstance(code, int), 'Provided exit code {} is not an int'.format(code)

    to_check = resolveProcesses(
        info_obj=proc_info,
        process=process,
        cmd_args=cmd_args,
        strict_proc_matching=strict_proc_matching
    )

    for info in [proc_info[item] for item in to_check]:
        assert info.returncode in allowable_exit_codes, 'Proc {} exited with code {}'.format(
            info.process_name,
            info.returncode
        )



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import executable as __executable

from launch.actions import ExecuteProcess as __ExecuteProcess

from .proc_lookup import NO_CMD_ARGS
from .proc_lookup import NoMatchingProcessException
from .proc_lookup import resolveProcesses


def KeepAliveProc():
    """
    Generate a dummy launch.actions.ExecuteProcess to keep the launch alive.

    launch_test expects to shut down the launch itself when it's done running tests.  If all
    of the processes under test are expected to terminate on their own, it's necessary to add
    another process to keep the launch service alive while the tests are running.
    """
    script = """
import signal
import time

try:
    while True:
        time.sleep(1)
except KeyboardInterrupt:
    pass
"""
    return __ExecuteProcess(
        cmd=[
            __executable,
            '-c',
            script
        ],
    )


__all__ = [
    'resolveProcesses',

    'KeepAliveProc',
    'NoMatchingProcessException',

    'NO_CMD_ARGS',
]



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import unittest

import ament_index_python

import launch
import launch.actions

import launch_testing
import launch_testing.actions
import launch_testing.asserts
import launch_testing.markers
import launch_testing.tools

import pytest


def get_test_process_action(*, args=[]):
    test_proc_path = os.path.join(
        ament_index_python.get_package_prefix('launch_testing'),
        'lib/launch_testing',
        'terminating_proc.py'
    )
    return launch.actions.ExecuteProcess(
        cmd=[sys.executable, test_proc_path, *args],
        name='terminating_proc.py',
        # This is necessary to get unbuffered output from the process under test
        additional_env={'PYTHONUNBUFFERED': '1'},
        output='screen'
    )


@pytest.mark.launch_test
@launch_testing.markers.keep_alive
def generate_test_description():
    return launch.LaunchDescription([launch_testing.actions.ReadyToTest()])


class TestTerminatingProc(unittest.TestCase):

    def test_no_args(self, launch_service, proc_info, proc_output):
        """Test terminating_proc without command line arguments."""
        proc_action = get_test_process_action()
        with launch_testing.tools.launch_process(
            launch_service, proc_action, proc_info, proc_output
        ):
            proc_info.assertWaitForStartup(process=proc_action, timeout=2)
            proc_output.assertWaitFor(
                'Starting Up', process=proc_action, timeout=2, stream='stdout'
            )
            proc_output.assertWaitFor(
                'Emulating Work', process=proc_action, timeout=2, stream='stdout'
            )
            proc_output.assertWaitFor('Done', process=proc_action, timeout=2, stream='stdout')
            proc_output.assertWaitFor(
                'Shutting Down', process=proc_action, timeout=2, stream='stdout'
            )
            proc_info.assertWaitForShutdown(process=proc_action, timeout=4)
        launch_testing.asserts.assertExitCodes(proc_info, process=proc_action)

    def test_with_args(self, launch_service, proc_info, proc_output):
        """Test terminating_proc with some command line arguments."""
        proc_action = get_test_process_action(args=['--foo', 'bar'])
        with launch_testing.tools.launch_process(
            launch_service, proc_action, proc_info, proc_output
        ):
            proc_info.assertWaitForStartup(process=proc_action, timeout=2)
            proc_output.assertWaitFor(
                'Starting Up', process=proc_action, timeout=2, stream='stdout'
            )
            proc_output.assertWaitFor(
                "Called with arguments ['--foo', 'bar']", process=proc_action, timeout=2,
                stream='stdout'
            )
            proc_output.assertWaitFor(
                'Emulating Work', process=proc_action, timeout=2, stream='stdout'
            )
            proc_output.assertWaitFor('Done', process=proc_action, timeout=2, stream='stdout')
            proc_output.assertWaitFor(
                'Shutting Down', process=proc_action, timeout=2, stream='stdout'
            )
            proc_info.assertWaitForShutdown(process=proc_action, timeout=4)
        launch_testing.asserts.assertExitCodes(proc_info, process=proc_action)

    def test_unhandled_exception(self, launch_service, proc_output, proc_info):
        """Test terminating_proc forcing an unhandled exception."""
        proc_action = get_test_process_action(args=['--exception'])
        with launch_testing.tools.launch_process(
            launch_service, proc_action, proc_info, proc_output
        ):
            proc_info.assertWaitForStartup(process=proc_action, timeout=2)
            proc_output.assertWaitFor(
                'Starting Up', process=proc_action, timeout=2, stream='stdout'
            )
            proc_output.assertWaitFor(
                "Called with arguments ['--exception']", process=proc_action, timeout=2,
                stream='stdout'
            )
            proc_info.assertWaitForShutdown(process=proc_action, timeout=4)
        launch_testing.asserts.assertExitCodes(
            proc_info, process=proc_action, allowable_exit_codes=[1]
        )



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def print_arguments_of_launch_description(*, launch_description):
    """Print the arguments of a LaunchDescription to the console."""
    print("Arguments (pass arguments as '<name>:=<value>'):")
    launch_arguments = launch_description.get_launch_arguments()
    any_conditional_arguments = False
    for argument_action in launch_arguments:
        msg = "\n    '"
        msg += argument_action.name
        msg += "':"
        if argument_action._conditionally_included:
            any_conditional_arguments = True
            msg += '*'
        msg += '\n        '
        msg += argument_action.description
        if argument_action.default_value is not None:
            default_str = ' + '.join([token.describe() for token in argument_action.default_value])
            msg += '\n        (default: {})'.format(default_str)
        print(msg)

    if len(launch_arguments) > 0:
        if any_conditional_arguments:
            print('\n* argument(s) which are only used if specific conditions occur')
    else:
        print('\n  No arguments.')



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import tools
from .decorator import post_shutdown_test
from .io_handler import ActiveIoHandler, IoHandler
from .parametrize import parametrize
from .proc_info_handler import ActiveProcInfoHandler, ProcInfoHandler
from .ready_aggregator import ReadyAggregator


__all__ = [
    # Modules
    'tools',

    # Functions
    'parametrize',
    'post_shutdown_test',

    # Classes
    'ActiveIoHandler',
    'ActiveProcInfoHandler',
    'IoHandler',
    'ProcInfoHandler',
    'ReadyAggregator',
]



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

from osrf_pycommon.terminal_color import remove_ansi_escape_sequences

from .text import build_line_match
from .text import build_text_match


def get_default_filtered_prefixes():
    return [
        'pid', 'rc',
    ]


def get_default_filtered_patterns():
    return []


def basic_output_filter(
    filtered_prefixes=None,
    filtered_patterns=None,
):
    """
    Create a line filtering function to help output testing.

    :param filtered_prefixes: A list of byte strings representing prefixes that will cause
    output lines to be ignored if they start with one of the prefixes. By default lines
    starting with the process ID (`'pid'`) and return code (`'rc'`) will be ignored.
    :param filtered_patterns: A list of byte strings representing regexes that will cause
    output lines to be ignored if they match one of the regexes.
    """
    if filtered_prefixes is None:
        filtered_prefixes = get_default_filtered_prefixes()
    if filtered_patterns is None:
        filtered_patterns = get_default_filtered_patterns()
    filtered_patterns = list(map(re.compile, filtered_patterns))

    def _filter(output):
        filtered_output_lines = []
        for line in output.splitlines(keepends=True):
            # Filter out stdout that comes from underlying DDS implementation
            # Note: we do not currently support matching filters across multiple stdout lines.
            if any(line.startswith(prefix) for prefix in filtered_prefixes):
                continue
            if any(pattern.match(line) for pattern in filtered_patterns):
                continue
            filtered_output_lines.append(line)
        return ''.join(filtered_output_lines)
    return _filter


def expected_output_from_file(path):
    """
    Get expected output lines from a file.

    :param path: path w/o extension of either a .txt file containing the lines
    to be matched or a .regex file containing patterns to be searched for.
    """
    literal_file = path + '.txt'
    if os.path.isfile(literal_file):
        with open(literal_file, 'r') as f:
            return f.read().splitlines()

    regex_file = path + '.regex'
    if os.path.isfile(regex_file):
        with open(regex_file, 'r') as f:
            return [re.compile(regex) for regex in f.read().splitlines()]

    raise RuntimeError('could not find output check file: {}'.format(path))


def expect_output(
    text=None,
    *,
    lines=None,
    expected_text=None,
    expected_lines=None,
    strip_ansi_escape_sequences=True,
    strict=False
):
    r"""
    Match output text or lines with expected text or lines.

    Either (expected) text or (expected) lines can be provided but giving both results
    in a ValueError.
    If lines are given but a text is expected, these lines are joined using '\n'.
    Likewise, if text is given but lines are expected, text is split into lines.

    :param expected_text: output text expectation, as supported
        by `launch_testing.tools.text.build_text_match`
    :param expected_lines: output lines expectation, as supported
        by `launch_testing.tools.text.build_line_match`
    :param text: output text to be matched
    :param lines: output text lines to be matched
    :param strip_ansi_escape_sequences: If True (default), strip
        ansi escape sequences from actual output before comparing
    """
    if (text is not None) == (lines is not None):
        raise ValueError('Either lines or text, but not both, must be specified')

    if (expected_text is not None) == (expected_lines is not None):
        raise ValueError('Either expected lines or text, but not both, must be specified')

    if expected_text is not None:
        if text is None:
            text = '\n'.join(lines)
        match = build_text_match(expected_text, strict=strict)
        if strip_ansi_escape_sequences:
            text = remove_ansi_escape_sequences(text)
        return match(text) is not None

    match = build_line_match(expected_lines, strict=strict)
    if lines is None:
        lines = text.splitlines()
    if strip_ansi_escape_sequences:
        lines = [remove_ansi_escape_sequences(line) for line in lines]
    return match(lines) is not None



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
import time
import unittest


def keep_alive(test_description):
    """Mark a test launch description to be kept alive after fixture processes' termination."""
    if not hasattr(test_description, '__markers__'):
        test_description.__markers__ = {}
    test_description.__markers__['keep_alive'] = True
    return test_description


def retry_on_failure(*, times, delay=None):
    """
    Mark a test case to be retried up to `times` on AssertionError.

    :param times: The number of times to rety the test.
    :param delay: The time to wait between retries, in seconds.
      A value of None will result in zero delay.
    """
    assert times > 0

    def _decorator(func):
        assert 'self' == list(inspect.signature(func).parameters)[0]

        @functools.wraps(func)
        def _wrapper(self, *args, **kwargs):
            n = times
            while n > 1:
                try:
                    ret = func(self, *args, **kwargs)
                    if isinstance(self, unittest.TestCase):
                        assert self._outcome.success
                    return ret
                except AssertionError:
                    self._outcome.errors.clear()
                    self._outcome.success = True
                    n -= 1
                if delay is not None:
                    time.sleep(delay)
            return func(self, *args, **kwargs)
        return _wrapper
    return _decorator



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module providing process IO capturing classes.

To prevent pytest from rewriting this module assertions, please PYTEST_DONT_REWRITE.
See https://docs.pytest.org/en/latest/assert.html#disabling-assert-rewriting for
further reference.
"""


import threading

from .asserts.assert_output import assertInStream
from .util import NoMatchingProcessException
from .util import resolveProcesses


class IoHandler:
    """
    Holds stdout captured from running processes.

    This class provides helper methods to enumerate the captured IO by individual processes
    """

    def __init__(self):
        self._sequence_list = []  # A time-ordered list of IO from all processes
        self._process_name_dict = {}  # A dict of time ordered lists of IO key'd by the process

    def track(self, process_name):
        if process_name not in self._process_name_dict:
            self._process_name_dict[process_name] = []

    def append(self, process_io):
        self._sequence_list.append(process_io)
        if process_io.process_name not in self._process_name_dict:
            self._process_name_dict[process_io.process_name] = []
        self._process_name_dict[process_io.process_name].append(process_io)

    def __iter__(self):
        return self._sequence_list.__iter__()

    def processes(self):
        """
        Get an iterable of unique launch.events.process.RunningProcessEvent objects.

        :returns [launch.actions.ExecuteProcess]:
        """
        return [val[0].action for val in self._process_name_dict.values() if len(val) > 0]

    def process_names(self):
        """
        Get the name of all unique processes that generated IO.

        :returns [string]:
        """
        return self._process_name_dict.keys()

    def __getitem__(self, key):
        """
        Get the output for a given process or process name.

        :param key: The process to get the output for
        :type key: String, or launch.actions.ExecuteProcess
        """
        if isinstance(key, str):
            return list(self._process_name_dict[key])
        else:
            return list(self._process_name_dict[key.process_details['name']])


class ActiveIoHandler:
    """
    Holds stdout captured from running processes.

    The ActiveIoHandler is meant to be used when capturing is still in progress and provides
    additional synchronization, as well as methods to wait on incoming IO
    """

    def __init__(self):
        self._sync_lock = threading.Condition()
        # Deliberately not calling the 'super' constructor here.  We're building this class
        # by composition so we can still give out the unsynchronized version
        self._io_handler = IoHandler()

    @property
    def io_event(self):
        return self._sync_lock

    def track(self, process_name):
        with self._sync_lock:
            self._io_handler.track(process_name)
            self._sync_lock.notify()

    def append(self, process_io):
        with self._sync_lock:
            self._io_handler.append(process_io)
            self._sync_lock.notify()

    def __iter__(self):
        with self._sync_lock:
            return list(self._io_handler).__iter__()

    def processes(self):
        """
        Get an iterable of unique launch.events.process.RunningProcessEvent objects.

        :returns [launch.actions.ExecuteProcess]:
        """
        with self._sync_lock:
            return list(self._io_handler.processes())

    def process_names(self):
        """
        Get the name of all unique processes that generated IO.

        :returns [string]:
        """
        with self._sync_lock:
            return list(self._io_handler.process_names())

    def __getitem__(self, key):
        """
        Get the output for a given process or process name.

        :param key: The process to get the output for
        :type key: String, or launch.actions.ExecuteProcess
        """
        with self._sync_lock:
            return self._io_handler[key]

    def assertWaitFor(self, *args, **kwargs):
        success = self.waitFor(*args, **kwargs)
        assert success, 'Waiting for output timed out'

    def waitFor(
        self,
        expected_output,
        process=None,  # Will wait for IO from all procs by default
        cmd_args=None,
        *,
        strict_proc_matching=True,
        output_filter=None,
        timeout=10,
        strip_ansi_escape_sequences=True,
        stream='stderr',
    ):
        success = False

        def msg_found():
            try:
                assertInStream(
                    self._io_handler,  # Use unsynchronized, since this is called from a lock
                    expected_output=expected_output,
                    process=process,
                    cmd_args=cmd_args,
                    output_filter=output_filter,
                    strict_proc_matching=strict_proc_matching,
                    strip_ansi_escape_sequences=strip_ansi_escape_sequences,
                    stream=stream,
                )
                return True
            except NoMatchingProcessException:
                # This can happen if no processes have generated any output yet.  It's not fatal.
                return False
            except AssertionError:
                return False

        with self._sync_lock:
            # TODO(pete.baughman): Searching through all of the IO can be time consuming/wasteful.
            # We can optimize this by making a note of where we left off searching and only
            # searching new messages when we return from the wait.
            success = self._sync_lock.wait_for(
                msg_found,
                timeout=timeout
            )

        if not success:
            # Help the user a little.  It's possible that they gave us a bad process name and no
            # had no hope of matching anything.
            matches = resolveProcesses(self,
                                       process=process,
                                       cmd_args=cmd_args,
                                       strict_proc_matching=False)
            if len(matches) == 0:
                raise Exception(
                    "After timeout, found no processes matching '{}'  "
                    "It either doesn't exist, was never launched, "
                    "or didn't generate any output".format(
                        process
                    )
                )

        return success



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import re

import ament_index_python


def get_default_filtered_prefixes():
    return [
        b'pid', b'rc',
    ]


def get_default_filtered_patterns():
    return []


def get_rmw_output_filter(rmw_implementation, filter_type):
    supported_filter_types = ['prefixes', 'patterns']
    if filter_type not in supported_filter_types:
        raise TypeError(
            'Unsupported filter_type "{0}". Supported types: {1}'
            .format(filter_type, supported_filter_types))
    resource_name = 'rmw_output_' + filter_type
    prefix_with_resource = ament_index_python.has_resource(
        resource_name, rmw_implementation)
    if not prefix_with_resource:
        return []

    # Treat each line of the resource as an independent filter.
    rmw_output_filter, _ = ament_index_python.get_resource(resource_name, rmw_implementation)
    return [str.encode(line) for line in rmw_output_filter.splitlines()]


def create_output_lines_filter(
    filtered_prefixes=None,
    filtered_patterns=None,
    filtered_rmw_implementation=None
):
    """
    Create a line filtering function to help output testing.

    :param filtered_prefixes: A list of byte strings representing prefixes that will cause
    output lines to be ignored if they start with one of the prefixes. By default lines
    starting with the process ID (`b'pid'`) and return code (`b'rc'`) will be ignored.
    :param filtered_patterns: A list of byte strings representing regexes that will cause
    output lines to be ignored if they match one of the regexes.
    :param filtered_rmw_implementation: RMW implementation for which the output will be
    ignored in addition to the `filtered_prefixes`/`filtered_patterns`.
    """
    filtered_prefixes = filtered_prefixes or get_default_filtered_prefixes()
    filtered_patterns = filtered_patterns or get_default_filtered_patterns()
    if filtered_rmw_implementation:
        filtered_prefixes.extend(get_rmw_output_filter(
            filtered_rmw_implementation, 'prefixes'
        ))
        filtered_patterns.extend(get_rmw_output_filter(
            filtered_rmw_implementation, 'patterns'
        ))
    filtered_patterns = map(re.compile, filtered_patterns)
    encoded_line_sep = os.linesep.encode('ascii')

    def _filter(output):
        filtered_output = []
        for line in output.splitlines():
            # Filter out stdout that comes from underlying DDS implementation
            # Note: we do not currently support matching filters across multiple stdout lines.
            if any(line.startswith(prefix) for prefix in filtered_prefixes):
                continue
            if any(pattern.match(line) for pattern in filtered_patterns):
                continue
            filtered_output.append(line)
        if output.endswith(encoded_line_sep):
            filtered_output.append(encoded_line_sep)
        return encoded_line_sep.join(filtered_output)
    return _filter


def create_output_lines_test(expected_lines):
    """Create output test given a list of expected lines."""
    def _collate(output, addendum):
        output.write(addendum)
        return output

    def _match(output, pattern):
        return any(pattern in line for line in output.getvalue().splitlines())

    return io.BytesIO(), _collate, _match, expected_lines


def create_output_regex_test(expected_patterns):
    """Create output test given a list of expected matching regular expressions."""
    def _collate(output, addendum):
        output.write(b'\n'.join(addendum.splitlines()))
        return output

    def _match(output, pattern):
        return pattern.search(output.getvalue()) is not None

    return io.BytesIO(), _collate, _match, expected_patterns


def create_output_test_from_file(output_file):
    """
    Create output test using the given file content.

    :param output_file: basename (i.e. w/o extension) of either a .txt file containing the
    lines to be matched or a .regex file containing patterns to be searched for.
    """
    literal_file = output_file + '.txt'
    if os.path.isfile(literal_file):
        with open(literal_file, 'rb') as f:
            expected_output = f.read().splitlines()
        return create_output_lines_test(expected_output)

    regex_file = output_file + '.regex'
    if os.path.isfile(regex_file):
        with open(regex_file, 'rb') as f:
            patterns = [re.compile(regex) for regex in f.read().splitlines()]
        return create_output_regex_test(patterns)

    raise RuntimeError('could not find output check file: {}'.format(output_file))



# Copyright 2019 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

import os


def normalize_lineseps(lines):
    r"""Normalize and then return the given lines to all use '\n'."""
    lines = lines.replace(os.linesep, '\n')
    # This happens (even on Linux and macOS) when capturing I/O from an
    # emulated tty.
    lines = lines.replace('\r\n', '\n')
    return lines


def build_line_match(expected_lines, *, strict=False):
    """
    Make a callable to match lines with ``expected_lines``.

    :param expected_lines: line expectation to match against
    :type expected_text: str or List[str] or List[Union[str, regex pattern]]
    :return: a callable that matches text against the expectation.
    :rtype: Callable[bool, [str]]
    """
    if isinstance(expected_lines, str):
        def _match(actual_lines, start=0):
            if strict:
                if all(expected_lines == line for line in actual_lines[start:]):
                    return start, start + len(actual_lines[start:])
                return None
            return next((
                (i, i + 1) for i, line in enumerate(
                    actual_lines[start:], start=start
                ) if expected_lines in line
            ), None)
        return _match

    if hasattr(expected_lines, 'match'):
        def _match(actual_lines, start):
            if strict:
                if all(expected_lines.match(line) for line in actual_lines[start:]):
                    return start, start + len(actual_lines[start:])
                return None
            return next((
                (i, i + 1) for i, line in enumerate(
                    actual_lines[start:], start=start
                ) if expected_lines.match(line)
            ), None)
        return _match

    if isinstance(expected_lines, Iterable):
        head_match, *tail_matches = [
            build_line_match(line, strict=False) for line in expected_lines
        ]

        def _match(actual_lines, start=0):
            next_start, end = head_match(actual_lines, start) or (-1, -1)
            if next_start < start or (strict and next_start != start):
                return None
            start = next_start
            for match in tail_matches:
                next_start, next_end = match(actual_lines, end) or (-1, -1)
                if next_start < end or (strict and next_start != end):
                    return None
                end = next_end
            return start, end
        return _match

    raise ValueError('Unknown format for expected lines')


def build_text_match(expected_text, *, strict=False):
    """
    Make a callable to match text with ``expected_text``.

    :param expected_text: text expectation to match against
    :type expected_text: str or regex pattern or List[Union[str, regex pattern]]
    :return: a callable that matches text against the expectation.
    :rtype: Callable[bool, [str]]
    """
    if isinstance(expected_text, str):
        def _match(actual_text, start=0):
            actual_text = normalize_lineseps(actual_text)
            if strict:
                if actual_text[start:] != expected_text:
                    return None
            else:
                start = actual_text.find(expected_text, start)
                if start < 0:
                    return None
            return start, start + len(expected_text)
        return _match

    if hasattr(expected_text, 'search') and hasattr(expected_text, 'match'):
        def _match(actual_text, start=0):
            actual_text = normalize_lineseps(actual_text)
            if strict:
                match = expected_text.match(actual_text, start)
            else:
                match = expected_text.search(actual_text, start)
            if match is not None:
                match = match.start(), match.end()
            return match
        return _match

    if isinstance(expected_text, Iterable):
        head_match, *tail_matches = [
            build_text_match(text, strict=False) for text in expected_text
        ]

        def _match(actual_text, start=0):
            actual_text = normalize_lineseps(actual_text)
            next_start, end = head_match(actual_text, start) or (-1, -1)
            if next_start < start or (strict and next_start != start):
                return None
            start = next_start
            for match in tail_matches:
                next_start, next_end = match(actual_text, end) or (-1, -1)
                if next_start < end or (strict and next_start != end):
                    return None
                end = next_end
            return start, end
        return _match

    raise ValueError('Unknown format for expected text')



# Copyright 2018 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from typing import List
from typing import Text
from typing import Tuple


# This was copy/pasted from ros2launch.api to avoid a rclpy dependency in launch_testing
def parse_launch_arguments(launch_arguments: List[Text]) -> List[Tuple[Text, Text]]:
    """Parse the given launch arguments from the command line, into list of tuples for launch."""
    parsed_launch_arguments = OrderedDict()  # type: ignore
    for argument in launch_arguments:
        count = argument.count(':=')
        if count == 0 or argument.startswith(':=') or (count == 1 and argument.endswith(':=')):
            raise RuntimeError(
                "malformed launch argument '{}', expected format '<name>:=<value>'"
                .format(argument))
        name, value = argument.split(':=', maxsplit=1)
        parsed_launch_arguments[name] = value  # last one wins is intentional
    return parsed_launch_arguments.items()



# Copyright 2019 Apex.AI, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .assert_exit_codes import assertExitCodes
from .assert_exit_codes import EXIT_OK
from .assert_exit_codes import EXIT_SIGINT
from .assert_exit_codes import EXIT_SIGKILL
from .assert_exit_codes import EXIT_SIGQUIT
from .assert_exit_codes import EXIT_SIGSEGV
from .assert_output import assertDefaultStream
from .assert_output import assertInStderr
from .assert_output import assertInStdout
from .assert_output import assertInStream
from .assert_sequential_output import assertSequentialStdout
from .assert_sequential_output import SequentialTextChecker

from ..util.proc_lookup import NO_CMD_ARGS

__all__ = [
    'assertDefaultStream',
    'assertExitCodes',
    'assertInStderr',
    'assertInStdout',
    'assertInStream',
    'assertSequentialStdout',

    'SequentialTextChecker',

    'NO_CMD_ARGS',

    'EXIT_OK',
    'EXIT_SIGINT',
    'EXIT_SIGKILL',
    'EXIT_SIGQUIT',
    'EXIT_SIGSEGV',
]
